<?php

namespace Ulmus\Adapter;

use Ulmus\ConnectionAdapter;
use Ulmus\Entity\Mysql\Table;
use Ulmus\Migration\FieldDefinition;
use Ulmus\Migration\MigrateInterface;
use Ulmus\QueryBuilder\Sql;
use Ulmus\Common\PdoObject;

use Ulmus\Exception\AdapterConfigurationException;
use Ulmus\Repository;

class MySQL implements AdapterInterface, MigrateInterface, SqlAdapterInterface {
    use SqlAdapterTrait;

    const ALLOWED_ATTRIBUTES = [
        'default', 'primary_key', 'auto_increment', 'update',
    ];

    const DSN_PREFIX = "mysql";

    public string $hostname;

    public string $database;

    protected string $username;

    protected string $password;

    public string $charset = "utf8mb4";

    public ? string $socket;
    
    public int $port = 3306;

    public function __construct(
        ?string $hostname = null,
        ?string $database = null,
        #[\SensitiveParameter]
        ?string $username = null,
        #[\SensitiveParameter]
        ?string $password = null,
        ?int $port = null,
        ?string $charset = null
    ) {
        if ($hostname) {
            $this->hostname = $hostname;
        }

        if ($database) {
            $this->database = $database;
        }

        if ($port) {
            $this->port = $port;
        }

        if ($username) {
            $this->username = $username;
        }

        if ($password) {
            $this->password = $password;
        }

        if ($charset) {
            $this->charset = $charset;
        }
    }

    public function connect() : PdoObject
    {
        try {
            $pdo = new PdoObject($this->buildDataSourceName(), $this->username, $this->password);
            $pdo->setAttribute(\PDO::ATTR_ERRMODE, \PDO::ERRMODE_EXCEPTION);
            $pdo->setAttribute(\PDO::ATTR_EMULATE_PREPARES, false);
            $pdo->setAttribute(\PDO::ATTR_AUTOCOMMIT, false);
            $pdo->setAttribute(\PDO::ATTR_DEFAULT_FETCH_MODE, \PDO::FETCH_ASSOC);
            $pdo->setAttribute(\PDO::MYSQL_ATTR_USE_BUFFERED_QUERY, false);
        }
        catch(\PDOException $ex){
            throw $ex;
        }
        finally {
            $this->password = str_repeat('*', random_int(8,16));
        }

        return $pdo;
    }
    
    public function buildDataSourceName() : string
    {
        if ( false !== ($this->socket ?? false) ) {
            $parts[] = "unix_socket={$this->socket}";            
        }
        else {
            $parts[] = "host={$this->hostname}";
            
            if ( false !== ( $this->port ?? false ) ) {            
                $parts[] = "port={$this->port}";
            }
        }
        
        $parts[] = "dbname={$this->database}";

        if ( $this->charset ?? false ) {
            $parts[] = "charset={$this->charset}";
        }

        return static::DSN_PREFIX . ":" . implode(';', $parts);
    }
    
    public function setup(array $configuration) : void
    {
        $connection = array_change_key_case($configuration, \CASE_LOWER);
        
        # Either Unix Socket is used, or Host / Port
        if ( null === ( $this->socket = $connection['socket'] ?? null ) ) {
            if ( "" === ( $this->hostname = $connection['host'] ?? "" ) ) {
                throw new AdapterConfigurationException("Your `host` name is missing from your configuration array");
            }
            
            if ( false !== ( $configuration['port'] ?? false ) ) {
                $this->port = $configuration['port'];
            }
        }
        
        if ( false === ( $this->database = $connection['database'] ?? false ) ) {
            throw new AdapterConfigurationException("Your `database` name is missing from your configuration array");
        }
        elseif ( false === ( $this->username = $connection['username'] ?? false ) ) {
            throw new AdapterConfigurationException("Your `username` is missing from your configuration array");
        }
        elseif ( false === ( $this->password = $connection['password'] ?? false ) ) {
            throw new AdapterConfigurationException("Your `password` is missing from your configuration array");
        }
        
        if ( false !== ( $configuration['charset'] ?? false ) ) {
            $this->charset = $configuration['charset'];
        }
    }

    public function mapFieldType(FieldDefinition $field, bool $typeOnly = false) : string
    {
        $mapper = new MySQLFieldMapper($field);

        return $typeOnly ? $mapper->type : $mapper->render();
    }

    public static function escapeIdentifier(string $segment, int $type) : string
    {
        switch($type) {
            case static::IDENTIFIER_DATABASE:
            case static::IDENTIFIER_TABLE:
            case static::IDENTIFIER_FIELD:
                return "`" . trim(str_replace("`", "``", $segment), '`') . "`";

            case static::IDENTIFIER_VALUE:
                return "\"$segment\"";
        }
    }

    public function defaultEngine(): ? string
    {
        return "InnoDB";
    }

    public function queryBuilderClass() : string
    {
        return Sql\MysqlQueryBuilder::class;
    }

    public function schemaTable(ConnectionAdapter $adapter, $databaseName, string $tableName) : null|object
    {
        return Table::repository(Repository::DEFAULT_ALIAS, $adapter)
            ->select(\Ulmus\Common\Sql::raw('this.*'))
            ->where($this->escapeIdentifier('table_schema', AdapterInterface::IDENTIFIER_FIELD), $databaseName)
            ->loadOneFromField($this->escapeIdentifier('table_name', AdapterInterface::IDENTIFIER_FIELD), $tableName);
    }
}
